import uqbar.containers


class ModuleNode(uqbar.containers.UniqueTreeNode):
    """
    A :py:class:`~uqbar.containers.UniqueTreeNode` subclass used during API
    construction to proxy modules.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Internals"

    ### INITIALIZER ###

    def __init__(self, name=None, documenter=None, source_path=None):
        super().__init__(name=name)
        self._documenter = documenter
        self._source_path = source_path

    ### SPECIAL METHODS ###

    def __str__(self):
        return "{}".format(self.name)

    ### PUBLIC PROPERTIES ###

    @property
    def documenter(self):
        return self._documenter

    @documenter.setter
    def documenter(self, documenter):
        self._documenter = documenter

    @property
    def package_path(self):
        return self._name

    @property
    def source_path(self):
        return self._source_path

    @source_path.setter
    def source_path(self, source_path):
        self._source_path = source_path


class PackageNode(uqbar.containers.UniqueTreeList):
    """
    A :py:class:`~uqbar.containers.UniqueTreeList` subclass used during
    API construction to proxy packages.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Internals"

    ### INITIALIZER ###

    def __init__(self, children=None, name=None, documenter=None, source_path=None):
        super().__init__(children=children, name=name)
        self._source_path = source_path
        self._documenter = documenter

    ### SPECIAL METHODS ###

    def __str__(self):
        result = ["{}/".format(self.name)]
        for child in self:
            result.extend("    " + line for line in str(child).splitlines())
        return "\n".join(result)

    ### PUBLIC PROPERTIES ###

    @property
    def _node_class(self):
        import uqbar.apis

        return (uqbar.apis.ModuleNode, uqbar.apis.PackageNode)

    @property
    def documenter(self):
        return self._documenter

    @documenter.setter
    def documenter(self, documenter):
        self._documenter = documenter

    @property
    def package_path(self):
        return self._name

    @property
    def source_path(self):
        return self._source_path

    @source_path.setter
    def source_path(self, source_path):
        self._source_path = source_path



master_doc = "index"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "uqbar.sphinx.api",
    "uqbar.sphinx.inheritance",
]

html_static_path = ["_static"]

uqbar_api_source_paths = ["fake_package"]
uqbar_api_root_documenter_class = "uqbar.apis.SummarizingRootDocumenter"
uqbar_api_module_documenter_class = "uqbar.apis.SummarizingModuleDocumenter"
uqbar_api_member_documenter_classes = [
    "uqbar.apis.FunctionDocumenter",
    "uqbar.apis.SummarizingClassDocumenter",
]



import collections
import itertools


def flatten(iterable):
    for x in iterable:
        if isinstance(x, collections.Iterable):
            yield from flatten(x)
        else:
            yield x


def group_by_count(iterable, count):
    iterator = iter(iterable)
    while True:
        group = list(itertools.islice(iterator, count))
        if not group:
            return
        yield group


def nwise(iterable, n=2):
    iterables = itertools.tee(iterable, n)
    temp = []
    for idx, it in enumerate(iterables):
        it = itertools.islice(it, idx, None)
        temp.append(it)
    return zip(*temp)


def repeat_to_length(sequence, length):
    for i, x in enumerate(itertools.cycle(sequence)):
        if i >= length:
            break
        yield x


def zip_cyclic(*args):
    maximum_i = max(len(_) for _ in args) - 1
    cycles = [itertools.cycle(_) for _ in args]
    iterator = enumerate(zip(*cycles))
    for i, result in iterator:
        yield result
        if i == maximum_i:
            break



import collections
import copy
import threading
import typing
from typing import Optional

from ..iterables import nwise


class UniqueTreeNode:
    """
    A node in a "unique" tree.

    Unique tree nodes may have at most one parent and may appear only once in
    the tree.
    """

    ### CLASS VARIABLES ###

    _state_flag_names: typing.Tuple[str, ...] = ()

    ### INITIALIZER ###

    def __init__(self, name: Optional[str] = None) -> None:
        self._name = name
        self._parent = None

    ### PRIVATE METHODS ###

    def _cache_named_children(self):
        name_dictionary = {}
        if self.name is not None:
            name_dictionary.setdefault(self.name, set()).add(self)
        return name_dictionary

    @classmethod
    def _iterate_nodes(cls, nodes):
        for x in nodes:
            yield x
            if hasattr(x, "_children"):
                for y in cls._iterate_nodes(x):
                    yield y

    def _get_node_state_flags(self):
        state_flags = {}
        for name in self._state_flag_names:
            state_flags[name] = True
            for node in self.parentage:
                if not getattr(node, name):
                    state_flags[name] = False
                    break
        return state_flags

    def _mark_entire_tree_for_later_update(self):
        for node in self.parentage:
            for name in self._state_flag_names:
                setattr(node, name, False)

    def _remove_from_parent(self):
        if self._parent is not None and self in self._parent:
            self._parent._children.remove(self)
        self._parent = None

    def _remove_named_children_from_parentage(self, old_parent, name_dictionary):
        if old_parent is None or not name_dictionary:
            return
        for parent in old_parent.parentage:
            named_children = parent._named_children
            for name in name_dictionary:
                for node in name_dictionary[name]:
                    named_children[name].remove(node)
                if not named_children[name]:
                    del named_children[name]

    def _restore_named_children_to_parentage(self, new_parent, name_dictionary):
        if new_parent is None or not name_dictionary:
            return
        for parent in new_parent.parentage:
            named_children = parent._named_children
            for name in name_dictionary:
                if name in named_children:
                    named_children[name].update(name_dictionary[name])
                else:
                    named_children[name] = copy.copy(name_dictionary[name])

    def _set_parent(self, new_parent):
        old_parent = self._parent
        named_children = self._cache_named_children()
        self._remove_from_parent()
        self._remove_named_children_from_parentage(old_parent, named_children)
        self._parent = new_parent
        self._restore_named_children_to_parentage(new_parent, named_children)
        if new_parent is None:
            self._mark_entire_tree_for_later_update()

    ### PUBLIC PROPERTIES ###

    @property
    def depth(self):
        return len(self.parentage) - 1

    @property
    def graph_order(self):
        """
        Get graph-order tuple for node.

        ::

            >>> from uqbar.containers import UniqueTreeList, UniqueTreeNode
            >>> root_container = UniqueTreeList(name="root")
            >>> outer_container = UniqueTreeList(name="outer")
            >>> inner_container = UniqueTreeList(name="inner")
            >>> node_a = UniqueTreeNode(name="a")
            >>> node_b = UniqueTreeNode(name="b")
            >>> node_c = UniqueTreeNode(name="c")
            >>> node_d = UniqueTreeNode(name="d")
            >>> root_container.extend([node_a, outer_container])
            >>> outer_container.extend([inner_container, node_d])
            >>> inner_container.extend([node_b, node_c])

        ::

            >>> for node in root_container.depth_first():
            ...     print(node.name, node.graph_order)
            ...
            a (0,)
            outer (1,)
            inner (1, 0)
            b (1, 0, 0)
            c (1, 0, 1)
            d (1, 1)

        """
        graph_order = []
        for parent, child in nwise(reversed(self.parentage)):
            try:
                index = parent.index(child)
            except AttributeError:
                index = 0
            graph_order.append(index)
        return tuple(graph_order)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, expr):
        assert isinstance(expr, (str, type(None)))
        old_name = self._name
        for parent in self.parentage[1:]:
            named_children = parent._named_children
            if old_name is not None:
                named_children[old_name].remove(self)
                if not named_children[old_name]:
                    del named_children[old_name]
            if expr is not None:
                if expr not in named_children:
                    named_children[expr] = set([self])
                else:
                    named_children[expr].add(self)
        self._name = expr

    @property
    def parent(self):
        return self._parent

    @property
    def parentage(self):
        parentage = []
        node = self
        while node is not None:
            parentage.append(node)
            node = getattr(node, "parent", None)
        return tuple(parentage)

    @property
    def root(self):
        proper_parentage = self.parentage[1:]
        if not proper_parentage:
            return None
        return proper_parentage[-1]


class UniqueTreeContainer(UniqueTreeNode):
    ### INITIALIZER ###

    def __init__(self, children=None, name=None):
        UniqueTreeNode.__init__(self, name=name)
        self._named_children = {}

    ### SPECIAL METHODS ###

    def __contains__(self, expr):
        if isinstance(expr, str):
            return expr in self._named_children
        for x in self._children:
            if x is expr:
                return True
        return False

    def __iter__(self):
        for child in self._children:
            yield child

    def __len__(self):
        return len(self._children)

    ### PRIVATE METHODS ###

    def _cache_named_children(self):
        name_dictionary = super()._cache_named_children()
        if hasattr(self, "_named_children"):
            for name, children in self._named_children.items():
                name_dictionary[name] = copy.copy(children)
        return name_dictionary

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self):
        return UniqueTreeNode

    ### PUBLIC METHODS ###

    def depth_first(self, top_down=True, prototype=None):
        for child in tuple(self):
            if top_down:
                if not prototype or isinstance(child, prototype):
                    yield child
            if isinstance(child, UniqueTreeContainer):
                yield from child.depth_first(top_down=top_down, prototype=prototype)
            if not top_down:
                if not prototype or isinstance(child, prototype):
                    yield child

    def recurse(self, prototype=None):
        return self.depth_first(prototype=prototype)

    ### PUBLIC PROPERTIES ###

    @property
    def children(self):
        return tuple(self._children)


class UniqueTreeList(UniqueTreeContainer):
    """
    A list-like node in a "unique" tree.

    List nodes may contain zero or more other nodes.

    Unique tree nodes may have at most one parent and may appear only once in
    the tree.
    """

    ### INITIALIZER ###

    def __init__(self, children=None, name=None):
        super().__init__(name=name)
        self._children = []
        if children is not None:
            self[:] = children

    ### SPECIAL METHODS ###

    def __delitem__(self, i):
        if isinstance(i, str):
            children = tuple(self._named_children[i])
            for child in children:
                parent = child.parent
                del parent[parent.index(child)]
            return
        if isinstance(i, int):
            if i < 0:
                i = len(self) + i
            i = slice(i, i + 1)
        self.__setitem__(i, [])
        self._mark_entire_tree_for_later_update()

    def __getitem__(self, expr):
        if isinstance(expr, (int, slice)):
            return self._children[expr]
        elif isinstance(expr, str):
            result = sorted(self._named_children[expr], key=lambda x: x.graph_order)
            if len(result) == 1:
                return result[0]
            return result
        raise ValueError(expr)

    def __setitem__(self, i, new_items):
        if isinstance(i, int):
            new_items = self._prepare_setitem_single(new_items)
            start_index, stop_index, _ = slice(i, i + 1).indices(len(self))
        else:
            new_items = self._prepare_setitem_multiple(new_items)
            start_index, stop_index, _ = i.indices(len(self))
        old_items = self[start_index:stop_index]
        self._validate(new_items, old_items, start_index, stop_index)
        self._set_items(new_items, old_items, start_index, stop_index)
        self._mark_entire_tree_for_later_update()

    ### PRIVATE METHODS ###

    def _prepare_setitem_multiple(self, expr):
        return list(expr)

    def _prepare_setitem_single(self, expr):
        return [expr]

    def _set_items(self, new_items, old_items, start_index, stop_index):
        for old_item in old_items:
            old_item._set_parent(None)
        for new_item in new_items:
            new_item._set_parent(self)
        self._children.__setitem__(slice(start_index, start_index), new_items)

    def _validate(self, new_nodes, old_nodes, start_index, stop_index):
        parentage = self.parentage
        for new_node in new_nodes:
            if not isinstance(new_node, self._node_class):
                raise ValueError(f"Expected {self._node_class}, got {type(new_node)}")
            elif new_node in parentage:
                raise ValueError("Cannot set parent node as child.")

    ### PUBLIC METHODS ###

    def append(self, expr):
        self.__setitem__(slice(len(self), len(self)), [expr])

    def extend(self, expr):
        self.__setitem__(slice(len(self), len(self)), expr)

    def index(self, expr):
        for i, child in enumerate(self._children):
            if child is expr:
                return i
        message = "{!r} not in {!r}."
        message = message.format(expr, self)
        raise ValueError(message)

    def insert(self, i, expr):
        self.__setitem__(slice(i, i), [expr])

    def pop(self, i=-1):
        node = self[i]
        del self[i]
        return node

    def remove(self, node):
        i = self.index(node)
        del self[i]


class UniqueTreeTuple(UniqueTreeContainer):
    """
    A tuple-like node in a "unique" tree.
    """

    ### INITIALIZER ###

    def __init__(self, children=None, name=None):
        super().__init__(name=name)
        self._children = []
        self._mutate(slice(None), children or ())

    ### SPECIAL METHODS ###

    def __getitem__(self, expr):
        if isinstance(expr, (int, slice)):
            return self._children[expr]
        elif isinstance(expr, str):
            result = sorted(self._named_children[expr], key=lambda x: x.graph_order)
            if len(result) == 1:
                return result[0]
            return result
        raise ValueError(expr)

    ### PRIVATE METHODS ###

    def _mutate(self, i, new_items):
        if isinstance(i, int):
            new_items = self._prepare_setitem_single(new_items)
            start_index, stop_index, _ = slice(i, i + 1).indices(len(self))
        else:
            new_items = self._prepare_setitem_multiple(new_items)
            start_index, stop_index, _ = i.indices(len(self))
        old_items = self[start_index:stop_index]
        self._validate(new_items, old_items, start_index, stop_index)
        self._set_items(new_items, old_items, start_index, stop_index)
        self._mark_entire_tree_for_later_update()

    def _prepare_setitem_multiple(self, expr):
        return list(expr)

    def _prepare_setitem_single(self, expr):
        return [expr]

    def _set_items(self, new_items, old_items, start_index, stop_index):
        for old_item in old_items:
            old_item._set_parent(None)
        for new_item in new_items:
            new_item._set_parent(self)
        self._children.__setitem__(slice(start_index, start_index), new_items)

    def _validate(self, new_nodes, old_nodes, start_index, stop_index):
        parentage = self.parentage
        for new_node in new_nodes:
            if not isinstance(new_node, self._node_class):
                raise ValueError(f"Expected {self._node_class}, got {type(new_node)}")
            elif new_node in parentage:
                raise ValueError("Cannot set parent node as child.")

    ### PUBLIC METHODS ###

    def index(self, expr):
        for i, child in enumerate(self._children):
            if child is expr:
                return i
        message = "{!r} not in {!r}."
        message = message.format(expr, self)
        raise ValueError(message)


class UniqueTreeSet(UniqueTreeContainer):
    """
    A set-like node in a "unique" tree.
    """

    ### INITIALIZER ###

    def __init__(self, children=None, name=None):
        super().__init__(name=name)
        self._children = set()
        self._lock = threading.RLock()
        if children is not None:
            self.update(children)

    ### PRIVATE METHODS ###

    def _mutate(self, new_nodes, old_nodes):
        with self._lock:
            self._validate(new_nodes, old_nodes)
            self._update_parentage(new_nodes, old_nodes)
            self._mark_entire_tree_for_later_update()

    def _update_parentage(self, new_nodes, old_nodes):
        for old_node in old_nodes:
            old_node._set_parent(None)
        for new_node in new_nodes:
            new_node._set_parent(self)
        self._children.update(new_nodes)

    def _validate(self, new_nodes):
        parentage = self.parentage
        for new_node in new_nodes:
            if not isinstance(new_node, self._node_class):
                raise ValueError(f"Expected {self._node_class}, got {type(new_node)}")
            elif new_node in parentage:
                raise ValueError("Cannot set parent node as child.")

    ### PUBLIC METHODS ###

    def add(self, node):
        self._mutate(new_nodes=(node,), old_nodes=())

    def clear(self):
        self._mutate(new_nodes=(), old_nodes=self._children)

    def pop(self):
        with self._lock:
            node = self._children.pop()
            self._mutate(new_nodes=(), old_nodes=(node,))
        return node

    def remove(self, node):
        with self._lock:
            self._children.remove(node)
            self._mutate(new_nodes=(), old_nodes=(node,))

    def update(self, expr):
        with self._lock:
            self._mutate(new_nodes=expr, old_nodes=())


class UniqueTreeDict(UniqueTreeContainer):
    """
    A dict-like node in a "unique" tree.
    """

    ### INITIALIZER ###

    def __init__(self, children=None, name=None):
        super().__init__(name=name)
        self._children = collections.OrderedDict()
        self._children_reversed = {}
        self._lock = threading.RLock()
        if children is not None:
            self.update(children)

    ### SPECIAL METHODS ###

    def __contains__(self, key):
        return key in self._children

    def __delitem__(self, key):
        with self._lock:
            self._mutate([], [(key, self._children[key])])

    def __getitem__(self, key):
        return self._children[key]

    def __iter__(self):
        return iter(self._children)

    def __setitem__(self, key, value):
        self._mutate([(key, value)], [])

    ### PRIVATE METHODS ###

    def _mutate(self, new_items, old_items):
        with self._lock:
            self._validate(new_items, old_items)
            new_nodes, old_nodes = self._update_items(new_items, old_items)
            self._update_parentage(new_nodes, old_nodes)
            self._mark_entire_tree_for_later_update()

    def _update_items(self, new_items, old_items):
        new_nodes, old_nodes = set(), set()
        for old_key, old_node in old_items:
            old_nodes.add(old_node)
            self._children.pop(old_key)
            self._children_reversed.pop(old_node)
        for _, new_node in new_items:
            new_nodes.add(new_node)
        for new_key, new_node in new_items:
            if new_key in self._children:
                # pre-existing key
                old_node = self[new_key]
                if old_node not in new_nodes:
                    old_nodes.add(old_node)
                    self._children_reversed.pop(old_node)
            if new_node in self._children_reversed:
                # pre-existing node
                self._children.pop(self._children_reversed[new_node])
            self._children[new_key] = new_node
            self._children_reversed[new_node] = new_key
        return new_nodes, old_nodes

    def _update_parentage(self, new_nodes, old_nodes):
        for old_node in old_nodes:
            old_node._set_parent(None)
        for new_node in new_nodes:
            new_node._set_parent(self)

    def _validate(self, new_items, old_items):
        parentage = self.parentage
        for _, new_node in new_items:
            if not isinstance(new_node, self._node_class):
                raise ValueError(f"Expected {self._node_class}, got {type(new_node)}")
            elif new_node in parentage:
                raise ValueError("Cannot set parent node as child.")

    ### PUBLIC METHODS ###

    def clear(self):
        self._mutate([], list(self.items()))

    def depth_first(self, top_down=True):
        for child in tuple(self.values()):
            if top_down:
                yield child
            if isinstance(child, UniqueTreeContainer):
                yield from child.depth_first(top_down=top_down)
            if not top_down:
                yield child

    def get(self, key, default=None):
        return self._children.get(key, default)

    def items(self):
        return self._children.items()

    def keys(self):
        return self._children.keys()

    def pop(self, *args):
        with self._lock:
            value = self._children.pop(*args)
            self._children[args[0]] = value
            self._mutate([], [(args[0], value)])
            return value

    def recurse(self):
        return self.depth_first()

    def update(*args, **kwargs):
        self, *args = args
        if len(args) > 1:
            raise TypeError(f"update expected at most 1 arguments, got {len(args)}")
        new_items = {}
        if args:
            new_items.update(dict(args[0]))
        new_items.update(**kwargs)
        self._mutate(list(new_items.items()), [])

    def values(self):
        return self._children.values()



#!/usr/bin/env python
import pathlib

import setuptools


def read_version():
    root_path = pathlib.Path(__file__).parent
    version_path = root_path / "uqbar" / "_version.py"
    with version_path.open() as file_pointer:
        file_contents = file_pointer.read()
    local_dict = {}
    exec(file_contents, None, local_dict)
    return local_dict["__version__"]


if __name__ == "__main__":
    setuptools.setup(
        package_data={"uqbar": ["py.typed"]},
        packages=setuptools.find_packages(
            include=["uqbar", "uqbar.*"], exclude=["*__pycache__*"]
        ),
        version=read_version(),
    )



class GrandParent:
    """
    A grand parent.

    ::

        >>> print("I am a grand parent.")

    """

    def one(self):
        """
        Grand parent: one.

        ::

            >>> print("Grand parent: one.")

        """
        pass

    def two(self):
        """
        Grand parent: two.

        ::

            >>> print("Grand parent: two.")

        """
        pass

    def three(self):
        """
        Grand parent: three.

        ::

            >>> print("Grand parent: three.")

        ::

            >>> import uqbar.graphs
            >>> g = uqbar.graphs.Graph()
            >>> n1 = uqbar.graphs.Node()
            >>> n2 = uqbar.graphs.Node()
            >>> g.extend([n1, n2])
            >>> e = n1.attach(n2)

        ::

            >>> print(format(g, "graphviz"))

        ::

            >>> uqbar.graphs.Grapher(g)()  # doctest: +SKIP

        And that, my friends, is a graph.
        """
        pass


class Parent(GrandParent):
    """
    A parent.

    ::

        >>> print("I am a parent.")

    """

    def one(self):
        """
        Parent: one.

        ::

            >>> print("Parent: one.")

        """
        pass


class Uncle(GrandParent):
    # I don't have a docstring or any overridden methods.
    pass


class Child(Parent):
    """
    A child.

    ::

        >>> print("I am a child.")

    """

    def two(self):
        """
        Child: two.

        ::

            >>> print("Child: two.")

        """
        pass


class Outer:
    """
    An outer class.

    ::

        >>> print("I am an outer class.")

    """

    class Inner:
        """
        An inner class.

        ::

            >>> print("I am an inner class.")

        """

        def inner_method(self):
            """
            Inner: method.

            ::

                >>> print("Inner: method.")

            """
            pass

    def outer_method(self):
        """
        Outer: method.

        ::

            >>> print("Outer: method.")

        """
        pass


def just_a_function():
    """
    Just a function.

    ::

        >>> print("I am just a function.")

    """
    pass



import uqbar.objects
import uqbar.strings


class MyNonReprObject:
    def __init__(self, arg1, arg2, *var_args, foo=None, bar=None, **kwargs):
        self.arg1 = arg1
        self.arg2 = arg2
        self.var_args = var_args
        self.foo = foo
        self.bar = bar
        self.kwargs = kwargs


class MyReprObject(MyNonReprObject):
    def __repr__(self):
        return uqbar.objects.get_repr(self)


class MyKwargOnlyObject:
    def __init__(self, *, foo, bar, baz):
        self.foo = foo
        self.bar = bar
        self.baz = baz


def test_objects_get_repr_01():
    my_object = MyNonReprObject("a", "b", "c", "d", foo="x", quux=["y", "z"])
    assert uqbar.objects.get_repr(my_object) == uqbar.strings.normalize(
        """
        MyNonReprObject(
            'a',
            'b',
            'c',
            'd',
            foo='x',
            quux=['y', 'z'],
        )
        """
    )


def test_objects_get_repr_02():
    my_object = MyReprObject("a", "b", "c", "d", foo="x", quux=["y", "z"])
    assert uqbar.objects.get_repr(my_object) == uqbar.strings.normalize(
        """
        MyReprObject(
            'a',
            'b',
            'c',
            'd',
            foo='x',
            quux=['y', 'z'],
        )
        """
    )


def test_objects_get_repr_03():
    my_object = MyReprObject("a", "b", "c", "d", foo="x", quux=["y", "z"])
    assert repr(my_object) == uqbar.strings.normalize(
        """
        MyReprObject(
            'a',
            'b',
            'c',
            'd',
            foo='x',
            quux=['y', 'z'],
        )
        """
    )


def test_objects_get_repr_04():
    object_a = MyReprObject("a", "b", "c", "d", foo="x", quux=["y", "z"])
    object_b = MyReprObject("pqr", object_a, bar=set([1, 2, 3]))
    assert repr(object_b) == uqbar.strings.normalize(
        """
        MyReprObject(
            'pqr',
            MyReprObject(
                'a',
                'b',
                'c',
                'd',
                foo='x',
                quux=['y', 'z'],
            ),
            bar={1, 2, 3},
        )
    """
    )


def test_objects_get_repr_05():
    object_a = MyReprObject("a", "b", "c", "d", foo="x", quux=["y", "z"])
    object_b = MyReprObject("pqr", [object_a], bar=set([1, 2, 3]))
    assert repr(object_b) == uqbar.strings.normalize(
        """
        MyReprObject(
            'pqr',
            [
                MyReprObject(
                    'a',
                    'b',
                    'c',
                    'd',
                    foo='x',
                    quux=['y', 'z'],
                ),
            ],
            bar={1, 2, 3},
        )
    """
    )


def test_objects_get_repr_06():
    my_object = MyKwargOnlyObject(foo="a", bar="b", baz="c")
    assert uqbar.objects.get_repr(my_object) == uqbar.strings.normalize(
        """
        MyKwargOnlyObject(
            bar='b',
            baz='c',
            foo='a',
        )
        """
    )



import pytest

import uqbar.containers


def test___contains__():
    graph = uqbar.containers.DependencyGraph()
    assert "A" not in graph
    assert "B" not in graph
    graph.add("A")
    assert "A" in graph
    assert "B" not in graph
    graph.add("B", parent="A")
    assert "A" in graph
    assert "B" in graph
    graph.remove("A")
    assert "A" not in graph
    assert "B" in graph
    graph.remove("B")
    assert "A" not in graph
    assert "B" not in graph


def test___eq__():
    graph_one = uqbar.containers.DependencyGraph()
    graph_two = uqbar.containers.DependencyGraph()
    graph_three = uqbar.containers.DependencyGraph()
    assert graph_one == graph_two
    assert graph_one != "foo"
    graph_one.add("A", parent="B")
    graph_one.add("C", parent="B")
    assert graph_one != graph_two
    graph_two.add("C", parent="B")
    graph_two.add("A", parent="B")
    assert graph_one != graph_two
    graph_three.add("C", parent="B")
    graph_three.add("A", parent="B")
    assert graph_two == graph_three


def test___getitem__():
    graph = uqbar.containers.DependencyGraph()
    graph.add("B", parent="A")
    graph.add("C", parent="A")
    graph.add("D", parent="A")
    graph.add("D", parent="B")
    assert graph["A"] == (frozenset([]), frozenset(["B", "C", "D"]))
    assert graph["B"] == (frozenset(["A"]), frozenset(["D"]))
    assert graph["C"] == (frozenset(["A"]), frozenset([]))
    assert graph["D"] == (frozenset(["A", "B"]), frozenset([]))


def test___iter__():
    graph = uqbar.containers.DependencyGraph()
    graph.add("A")
    graph.add("B", parent="A")
    graph.add("C", parent="A")
    graph.add("D", parent="B")
    graph.add("D", parent="C")
    graph.add("E", parent="A")
    assert list(graph) == ["D", "B", "C", "E", "A"]
    assert list(graph) == ["D", "B", "C", "E", "A"]  # non-destructive
    graph.remove("A")
    assert list(graph) == ["D", "B", "C", "E"]
    graph.remove("C")
    assert list(graph) == ["D", "B", "E"]


def test___len__():
    graph = uqbar.containers.DependencyGraph()
    assert len(graph) == 0
    graph.add("A")
    assert len(graph) == 1
    graph.add("B", parent="A")
    assert len(graph) == 2
    graph.add("C", parent="A")
    assert len(graph) == 3
    graph.remove("A")
    assert len(graph) == 2
    graph.remove("B")
    assert len(graph) == 1
    graph.remove("C")
    assert len(graph) == 0


def test_add():
    graph = uqbar.containers.DependencyGraph()
    graph.add("A")
    assert "A" in graph
    graph.add("B", parent="C")
    assert "B" in graph
    assert "C" in graph
    graph.add("B", parent="A")
    assert "A" in graph
    assert "B" in graph
    assert "C" in graph


def test_children():
    graph = uqbar.containers.DependencyGraph()
    graph.add("B", parent="A")
    graph.add("C", parent="A")
    graph.add("D")
    graph.add("E", parent="A")
    graph.add("E", parent="D")
    assert graph.children("A") == frozenset(["B", "C", "E"])
    assert graph.children("B") == frozenset([])
    assert graph.children("C") == frozenset([])
    assert graph.children("D") == frozenset(["E"])
    assert graph.children("E") == frozenset([])


def test_copy():
    graph_one = uqbar.containers.DependencyGraph()
    graph_one.add("A", parent="B")
    graph_one.add("C", parent="B")
    graph_two = graph_one.copy()
    assert graph_one == graph_two
    assert graph_one is not graph_two
    graph_one.remove("B")
    assert graph_one != graph_two


def test_is_acyclic():
    graph = uqbar.containers.DependencyGraph()
    assert graph.is_acyclic()
    graph.add("A")
    assert graph.is_acyclic()
    graph.add("B", parent="A")
    assert graph.is_acyclic()
    graph.add("C", parent="B")
    assert graph.is_acyclic()
    graph.add("A", parent="C")
    assert not graph.is_acyclic()
    graph.remove("B")
    assert graph.is_acyclic()


def test_parents():
    graph = uqbar.containers.DependencyGraph()
    graph.add("B", parent="A")
    graph.add("C", parent="A")
    graph.add("D")
    graph.add("E", parent="A")
    graph.add("E", parent="D")
    assert graph.parents("A") == frozenset([])
    assert graph.parents("B") == frozenset(["A"])
    assert graph.parents("C") == frozenset(["A"])
    assert graph.parents("D") == frozenset([])
    assert graph.parents("E") == frozenset(["A", "D"])


def test_remove():
    graph = uqbar.containers.DependencyGraph()
    with pytest.raises(ValueError):
        graph.remove("A")
    assert "A" not in graph
    graph.add("A")
    assert "A" in graph
    graph.remove("A")
    assert "A" not in graph



import collections
import inspect
from typing import Any, Dict, Optional, TypeVar

T = TypeVar("T")


def _dispatch_formatting(expr):
    if isinstance(expr, (list, tuple)):
        return _get_sequence_repr(expr)
    return repr(expr)


def _get_object_signature(expr):
    return inspect.signature(type(expr).__init__)


def _get_sequence_repr(expr):
    prototype = (bool, int, float, str, type(None))
    if all(isinstance(x, prototype) for x in expr):
        result = repr(expr)
        if len(result) < 50:
            return result
    if isinstance(expr, list):
        braces = "[", "]"
    else:
        braces = "(", ")"
    result = [braces[0]]
    for x in expr:
        for line in repr(x).splitlines():
            result.append("    " + line)
        result[-1] += ","
    result.append(braces[-1])
    return "\n".join(result)


def compare_objects(object_one, object_two, coerce=False):
    if coerce:
        try:
            object_two = type(object_one)(object_two)
        except (ValueError, TypeError):
            return False
    object_one_values = type(object_one), get_vars(object_one)
    try:
        object_two_values = type(object_two), get_vars(object_two)
    except AttributeError:
        object_two_values = type(object_two), object_two
    return object_one_values == object_two_values


def get_hash(expr):
    args, var_args, kwargs = get_vars(expr)
    hash_values = [type(expr)]
    for key, value in args.items():
        if isinstance(value, list):
            value = tuple(value)
        elif isinstance(value, set):
            value = frozenset(value)
        elif isinstance(value, dict):
            value = tuple(sorted(value.items()))
        args[key] = value
    hash_values.append(tuple(args.items()))
    hash_values.append(tuple(var_args))
    for key, value in kwargs.items():
        if isinstance(value, list):
            value = tuple(value)
        elif isinstance(value, set):
            value = frozenset(value)
        elif isinstance(value, dict):
            value = tuple(sorted(value.items()))
        kwargs[key] = value
    hash_values.append(tuple(sorted(kwargs.items())))
    return hash(tuple(hash_values))


def get_repr(
    expr, multiline: Optional[bool] = None, suppress_defaults: bool = True
) -> str:
    """
    Build a repr string for ``expr`` from its vars and signature.

    ::

        >>> class MyObject:
        ...     def __init__(self, arg1, arg2, *var_args, foo=None, bar=None, **kwargs):
        ...         self.arg1 = arg1
        ...         self.arg2 = arg2
        ...         self.var_args = var_args
        ...         self.foo = foo
        ...         self.bar = bar
        ...         self.kwargs = kwargs
        ...
        >>> my_object = MyObject('a', 'b', 'c', 'd', foo='x', quux=['y', 'z'])

    ::

        >>> import uqbar
        >>> print(uqbar.objects.get_repr(my_object))
        MyObject(
            'a',
            'b',
            'c',
            'd',
            foo='x',
            quux=['y', 'z'],
        )

    """
    signature = _get_object_signature(expr)
    if signature is None:
        return "{}()".format(type(expr).__name__)

    defaults = {}
    for name, parameter in signature.parameters.items():
        if parameter.default is not inspect._empty:
            defaults[name] = parameter.default

    args, var_args, kwargs = get_vars(expr)
    args_parts = collections.OrderedDict()
    var_args_parts = []
    kwargs_parts = {}
    has_lines = bool(multiline)
    parts = []

    # Format keyword-optional arguments.
    # print(type(expr), args)
    for i, (key, value) in enumerate(args.items()):
        arg_repr = _dispatch_formatting(value)
        if "\n" in arg_repr:
            has_lines = True
        args_parts[key] = arg_repr

    # Format *args
    for arg in var_args:
        arg_repr = _dispatch_formatting(arg)
        if "\n" in arg_repr:
            has_lines = True
        var_args_parts.append(arg_repr)

    # Format **kwargs
    for key, value in sorted(kwargs.items()):
        if suppress_defaults and key in defaults and value == defaults[key]:
            continue
        value = _dispatch_formatting(value)
        arg_repr = "{}={}".format(key, value)
        has_lines = True
        kwargs_parts[key] = arg_repr

    for _, part in args_parts.items():
        parts.append(part)
    parts.extend(var_args_parts)
    for _, part in sorted(kwargs_parts.items()):
        parts.append(part)

    # If we should format on multiple lines, add the appropriate formatting.
    if has_lines and parts and multiline is not False:
        for i, part in enumerate(parts):
            parts[i] = "\n".join("    " + line for line in part.split("\n"))
        parts.append(")")
        return "{}(\n{}".format(type(expr).__name__, ",\n".join(parts))

    return "{}({})".format(type(expr).__name__, ", ".join(parts))


def get_vars(expr):
    """
    Get ``args``, ``var args`` and ``kwargs`` for an object ``expr``.

    ::

        >>> class MyObject:
        ...     def __init__(self, arg1, arg2, *var_args, foo=None, bar=None, **kwargs):
        ...         self.arg1 = arg1
        ...         self.arg2 = arg2
        ...         self.var_args = var_args
        ...         self.foo = foo
        ...         self.bar = bar
        ...         self.kwargs = kwargs
        ...
        >>> my_object = MyObject('a', 'b', 'c', 'd', foo='x', quux=['y', 'z'])

    ::

        >>> import uqbar
        >>> args, var_args, kwargs = uqbar.objects.get_vars(my_object)

    ::

        >>> dict(args)
        {'arg1': 'a', 'arg2': 'b'}

    ::

        >>> var_args
        ['c', 'd']

    ::

        >>> kwargs
        {'foo': 'x', 'bar': None, 'quux': ['y', 'z']}

    """

    def _get_vars(signature):
        args = collections.OrderedDict()
        var_args = []
        kwargs = {}
        if expr is None:
            return args, var_args, kwargs
        for i, (name, parameter) in enumerate(signature.parameters.items()):
            # print('   ', parameter, parameter.kind, parameter.default)
            if i == 0 and name in ("self", "cls", "class_", "klass"):
                continue
            if parameter.kind is inspect._POSITIONAL_ONLY:
                try:
                    args[name] = getattr(expr, name)
                except AttributeError:
                    args[name] = expr[name]
            elif parameter.kind in (
                inspect._POSITIONAL_OR_KEYWORD,
                inspect._KEYWORD_ONLY,
            ):
                for x in (name, "_" + name):
                    try:
                        value = getattr(expr, x)
                        break
                    except AttributeError:
                        try:
                            value = expr[x]
                            break
                        except (KeyError, TypeError):
                            pass
                else:
                    raise ValueError("Cannot find value for {!r}".format(name))
                # print("        ", value)
                if parameter.kind is inspect._KEYWORD_ONLY:
                    # print("        ??? A")
                    kwargs[name] = value
                    continue
                elif parameter.default is inspect._empty:
                    # print("        ??? B")
                    args[name] = value
                    continue
                elif parameter.default == value:
                    # print("        ??? C")
                    kwargs[name] = value
                    continue
                # print("        ??? D")
                kwargs[name] = value
            elif parameter.kind is inspect._VAR_POSITIONAL:
                value = None
                try:
                    value = expr[:]
                except TypeError:
                    value = getattr(expr, name)
                if value:
                    var_args.extend(value)
            elif parameter.kind is inspect._VAR_KEYWORD:
                items = {}
                if hasattr(expr, "items"):
                    items = expr.items()
                elif hasattr(expr, name):
                    mapping = getattr(expr, name)
                    if not isinstance(mapping, dict):
                        mapping = dict(mapping)
                    items = mapping.items()
                elif hasattr(expr, "_" + name):
                    mapping = getattr(expr, "_" + name)
                    if not isinstance(mapping, dict):
                        mapping = dict(mapping)
                    items = mapping.items()
                for key, value in items:
                    if key not in args:
                        kwargs[key] = value
        return args, var_args, kwargs

    try:
        return _get_vars(inspect.signature(type(expr).__init__))
    except AttributeError:
        return _get_vars(inspect.signature(type(expr).__new__))


def new(expr: T, *args, **kwargs) -> T:
    """
    Template an object.

    ::

        >>> class MyObject:
        ...     def __init__(self, arg1, arg2, *var_args, foo=None, bar=None, **kwargs):
        ...         self.arg1 = arg1
        ...         self.arg2 = arg2
        ...         self.var_args = var_args
        ...         self.foo = foo
        ...         self.bar = bar
        ...         self.kwargs = kwargs
        ...
        >>> my_object = MyObject('a', 'b', 'c', 'd', foo='x', quux=['y', 'z'])

    ::

        >>> import uqbar
        >>> new_object = uqbar.objects.new(my_object, foo=666, bar=1234)
        >>> print(uqbar.objects.get_repr(new_object))
        MyObject(
            'a',
            'b',
            'c',
            'd',
            bar=1234,
            foo=666,
            quux=['y', 'z'],
        )

    Original object is unchanged:

    ::

        >>> print(uqbar.objects.get_repr(my_object))
        MyObject(
            'a',
            'b',
            'c',
            'd',
            foo='x',
            quux=['y', 'z'],
        )

    """
    # TODO: Clarify old vs. new variable naming here.
    current_args, current_var_args, current_kwargs = get_vars(expr)
    new_kwargs = current_kwargs.copy()

    recursive_arguments: Dict[str, Any] = {}
    for key in tuple(kwargs):
        if "__" in key:
            value = kwargs.pop(key)
            key, _, subkey = key.partition("__")
            recursive_arguments.setdefault(key, []).append((subkey, value))

    for key, pairs in recursive_arguments.items():
        recursed_object = current_args.get(key, current_kwargs.get(key))
        if recursed_object is None:
            continue
        kwargs[key] = new(recursed_object, **dict(pairs))

    if args:
        current_var_args = args
    for key, value in kwargs.items():
        if key in current_args:
            current_args[key] = value
        else:
            new_kwargs[key] = value

    new_args = list(current_args.values()) + list(current_var_args)
    return type(expr)(*new_args, **new_kwargs)
